﻿using System;

namespace mosaic {
    class Colour {
        private readonly int r, g, b;

        private double[] xyz = null;
        private double[] lab = null;

        public int R {
            get {
                return r;
            }
        }

        public int G {
            get {
                return g;
            }
        }

        public int B {
            get {
                return b;
            }
        }

        /// <summary>
        /// Ctor from a single 32-bit integer that stores the colour in RGBA format.
        /// Alpha value is ignored, as the application does not support transparency.
        /// </summary>
        /// <param name="rgba">Colour in RGBA format</param>
        public Colour(int rgba) {
            this.r = rgba >> 24 & 0xFF;
            this.g = (rgba >> 16) & 0xFF;
            this.b = (rgba >> 8) & 0xFF;

            xyz = GetXYZ();
            lab = GetLab();
        }

        /// <summary>
        /// Ctor from separate integers representing RGB channels.
        /// </summary>
        /// <param name="r">Red value 0-255</param>
        /// <param name="g">Green value 0-255</param>
        /// <param name="b">Blue value 0-255</param>
        public Colour(int r, int g, int b) {
            this.r = r;
            this.g = g;
            this.b = b;

            xyz = GetXYZ();
            lab = GetLab();
        }

        /// <summary>
        /// Calculates the distance of a colour from this one in the Lab colour space.
        /// </summary>
        /// <param name="c">The colour to compare this one with</param>
        /// <returns>The calculated distance</returns>
        public double GetDistanceFrom(Colour c) {
            double[] Lab1 = this.lab;
            double[] Lab2 = c.lab;

            return Math.Sqrt(
                (Lab2[0] - Lab1[0])*(Lab2[0] - Lab1[0]) +
                (Lab2[1] - Lab1[1])*(Lab2[1] - Lab1[1]) +
                (Lab2[2] - Lab1[2])*(Lab2[2] - Lab1[2])
                );
        }

        #region Black magic

        /// <summary>
        /// Calculates XYZ colour space values from the stored RGB values.
        /// </summary>
        /// <returns>A 3-length array with the XYZ values</returns>
        private double[] GetXYZ() {
            double _r = (double) r/255.0;
            double _g = (double) g/255.0;
            double _b = (double) b/255.0;

            if (_r > 0.04045) {
                _r = Math.Pow((_r + 0.055)/1.055, 2.4);
            }
            else {
                _r = _r/12.92;
            }

            if (_g > 0.04045) {
                _g = Math.Pow((_g + 0.055)/1.055, 2.4);
            }
            else {
                _g = _g/12.92;
            }

            if (_b > 0.04045) {
                _b = Math.Pow((_b + 0.055)/1.055, 2.4);
            }
            else {
                _b = _b/12.92;
            }

            _r *= 100;
            _g *= 100;
            _b *= 100;

            double x = .4124564*_r + .3575761*_g + .1804375*_b;
            double y = .2126729*_r + .7151522*_g + .0721750*_b;
            double z = .0193339*_r + .1191920*_g + .9503041*_b;

            return new double[3] {x, y, z};
        }

        /// <summary>
        /// Helper function for the GetLab function
        /// </summary>
        private double f(double t) {
            if (t > Math.Pow(6/29.0, 3)) {
                return Math.Pow(t, 1/3.0);
            }
            else {
                return 1/3.0*Math.Pow(29/6.0, 2)*t + 4/29.0;
            }
        }

        /// <summary>
        /// Calculates the colour's Lab colour space representation.
        /// </summary>
        /// <returns>The Lab values in a 3-length array</returns>
        private double[] GetLab() {
            double[] xyz = this.xyz ?? GetXYZ();

            double x = xyz[0];
            double y = xyz[1];
            double z = xyz[2];

            double L = 116*f(y/100.000) - 16;
            double a = 500*(f(x/95.047) - f(y/100.000));
            double b = 200*(f(y/100.000) - f(z/108.883));

            return new double[] {L, a, b};
        }

        #endregion
    }
}
