﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Windows.Forms;

namespace mosaic {
    public partial class MosaicDisplay : UserControl {
        private Bitmap bitmap;
        private Bitmap mosaic;
        private ColourManager cm;

        public MosaicDisplay() {
            InitializeComponent();
        }

        #region Public events

        public delegate void IntValueChanged(int newValue);

        public event IntValueChanged SquareSizeChanged;

        public delegate void DoubleValueChanged(double newValue);

        public event DoubleValueChanged ScaleChanged;

        #endregion

        #region Properties

        /// <summary>
        /// Checks if a value is in a given range and is unique from another given value
        /// </summary>
        /// <typeparam name="T">The type of the values</typeparam>
        /// <param name="newValue">The new value to check</param>
        /// <param name="oldValue">The old value from which the new one should be different</param>
        /// <param name="beginRange">The beginning of the given range</param>
        /// <param name="endRange">The end of the given range</param>
        /// <returns>True if newValue meets the requirements</returns>
        private static bool IsDifferentAndInRange<T>(T newValue, T oldValue, T beginRange, T endRange)
            where T : IComparable {
            return oldValue.CompareTo(newValue) != 0 &&
                   beginRange.CompareTo(newValue) <= 0 &&
                   endRange.CompareTo(newValue) >= 0;
        }

        private int squareSize = 12;

        /// <summary>
        /// The size of a square on the display in pixels
        /// </summary>
        public int Square {
            get {
                return squareSize;
            }
            set {
                if (IsDifferentAndInRange(value, squareSize, 1, 128)) {
                    squareSize = value;
                    RenderImage();

                    if (SquareSizeChanged != null) {
                        SquareSizeChanged(Square);
                    }
                }
            }
        }

        private double scale = 1.0;

        /// <summary>
        /// A scaling factor for the displayed squares
        /// </summary>
        public double Scale {
            get {
                return scale;
            }
            set {
                if (IsDifferentAndInRange(value, scale, 0.5, 5.0)) {
                    scale = value;

                    if (ScaleChanged != null) {
                        ScaleChanged(Scale);
                    }
                    UpdateBitmap();
                    Refresh();
                }
            }
        }

        #endregion

        #region Public control functions

        /// <summary>
        /// Opens a new image to display
        /// </summary>
        /// <param name="fileName">The image file's filename</param>
        public void OpenImage(string fileName) {
            if (cm == null) {
                cm = new ColourManager(new FileInfo("colours.txt"));
            }
            bitmap = new Bitmap(fileName);
            RenderImage();
        }

        /// <summary>
        /// Loads a new set of colours to use for display
        /// </summary>
        /// <param name="filename">The colour file's filename</param>
        public void LoadPalette(string filename) {
            if (filename == null) {
                filename = "colours.txt";
            }

            cm = new ColourManager(new FileInfo(filename));

            RenderImage();
        }

        private const double scaleStep = 1.0;

        /// <summary>
        /// Zooms in on the current image
        /// </summary>
        public void ZoomIn() {
            Scale += scaleStep;
        }

        /// <summary>
        /// Zooms out from the current image
        /// </summary>
        public void ZoomOut() {
            Scale -= scaleStep;
        }

        private const int squareSizeStep = 2;

        /// <summary>
        /// Increases the resolution of the image, i.e. makes the squares smaller
        /// </summary>
        public void IncreaseRes() {
            Square -= squareSizeStep;
        }

        /// <summary>
        /// Decreases the resolution of the image, i.e. makes the squares larger
        /// </summary>
        public void DecreaseRes() {
            Square += squareSizeStep;
        }

        #endregion

        #region Generation and display

        private List<MosaicSquare> squares = new List<MosaicSquare>();
        private RenderDialog renderDialog;

        /// <summary>
        /// Renders the currently opened image, displays a dialog while rendering
        /// </summary>
        private void RenderImage() {
            if (bitmap == null) {
                return;
            }

            renderDialog = new RenderDialog(RenderSquares);
            renderDialog.Show(this);
            renderDialog.Render();

            Refresh();
        }

        /// <summary>
        /// Empties the list of currently stored squares, and creates new ones
        /// using the currently open image and settings. Updates the displayed
        /// image to show the new squares.
        /// </summary>
        private void RenderSquares() {
            int hori_pieces = bitmap.Width/Square;
            int vert_pieces = bitmap.Height/Square;

            foreach (MosaicSquare ms in squares) {
                ms.Dispose();
            }
            squares.Clear();

            double totalPieces = hori_pieces*vert_pieces;
            int piecesDone = 0;
            renderDialog.ProgressValue = 0;

            for (int x = 0; x < hori_pieces; x++) {
                for (int y = 0; y < vert_pieces; y++) {
                    squares.Add(CreateSquare(x, y, Square));
                    piecesDone++;
                }
                renderDialog.ProgressValue = piecesDone/totalPieces;
            }

            UpdateBitmap();
        }

        /// <summary>
        /// Updates the size of display elements using the current
        /// square size and scale factor.
        /// </summary>
        private void UpdateSizes() {
            int width = (bitmap.Width/Square)*Square*(int) Scale;
            int height = (bitmap.Height/Square)*Square*(int) Scale;

            mosaic = new Bitmap(width, height);
            pictureBox1.Image = mosaic;

            pictureBox1.Width = width;
            pictureBox1.Height = height;
        }

        /// <summary>
        /// Paints the stored squares on the displayed bitmap.
        /// </summary>
        private void UpdateBitmap() {
            UpdateSizes();
            using (Graphics g = Graphics.FromImage(mosaic)) {
                foreach (MosaicSquare ms in squares) {
                    ms.Paint(g, Scale);
                }
                g.Flush();
            }
        }

        /// <summary>
        /// Creates a square for the given parameters, using the currently open image.
        /// </summary>
        /// <param name="x">The horizontal coordinate of the square in the grid</param>
        /// <param name="y">The vertical coordinate of the square in the grid</param>
        /// <param name="squareSize">The size of a square's side</param>
        /// <returns>The newly created square</returns>
        private MosaicSquare CreateSquare(int x, int y, int squareSize) {
            int r = 0, g = 0, b = 0;

            for (int i = 0; i < squareSize; i++) {
                for (int j = 0; j < squareSize; j++) {
                    Color pixel = bitmap.GetPixel(squareSize*x + i, squareSize*y + j);
                    r += pixel.R;
                    g += pixel.G;
                    b += pixel.B;
                }
            }

            int pixelCount = squareSize*squareSize;
            double _r = (double) r/pixelCount;
            double _g = (double) g/pixelCount;
            double _b = (double) b/pixelCount;

            Colour colour = new Colour((int) _r, (int) _g, (int) _b);

            return new MosaicSquare(cm.GetClosestStoredColour(colour), x*squareSize, y*squareSize, squareSize);
        }

        /// <summary>
        /// Returns the currently displayed image as a Bitmap
        /// </summary>
        /// <returns>The currently displayed mosaic</returns>
        public Bitmap GetBitMap() {
            int width = (bitmap.Width/Square)*Square;
            int height = (bitmap.Height/Square)*Square;
            Bitmap bmp = new Bitmap(width, height);

            using (Graphics g = Graphics.FromImage(bmp)) {
                foreach (MosaicSquare ms in squares) {
                    ms.Paint(g, 1.0);
                }
                g.Flush();
            }

            return bmp;
        }

        #endregion
    }
}
